"""
class Inputs(urwid.ListBox):
    def keypress(self, size, key):
        key = super(Inputs, self).keypress(size, key)
        if key == 


def text(program, handlers=None, freedesktop=False):
    '''
    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface,
    '''
    import urwid
    urwid.MainLoop(Main).run()

ListBox
    ListBoxRow
    ListBoxRow
    ListBoxRow
        Box
            Button
            Button
    ListBoxRow
       TextView
"""
